#!/usr/bin/env python
# coding: utf-8

"""
Tools to standardize notations.

The following abbreviations define the contexts:
    stm: the string is used in the stem of a file name
    ttl: the string is a title and contains LaTeX code
    csl: the string is displayed in a console
"""

from . import *

@beartype
def number(
    x: Scalar,
    c: str = 'csl',
    w: int = 4,
) -> str:
    """
    Return a notation of the number x in context c.

    Input:
        x (Scalar): number
        c (str): context
        w (int): width of the output string

    Output:
        s (str): notation of x
    """
    S = 0 if x>=0 else 1 # 0 for + / 1 for -
    m, e = f"{x:e}".split('e')
    m, e = eval(m), int(e) # mantissa and exponent
    A, B = f"{x:f}".rstrip('0').strip('-').split('.')
    a, b = len(A), len(B) # number of digits before / after the comma
    if isinstance(x, int) or x%1==0:
        if a+S<=w or a<=3:
            if c == 'stm':
                return f"{S*'-'+A.zfill(w-S)}"
            if c == 'ttl':
                return f"$ {S*'-'+A} $"
            if c == 'csl':
                return format(S*'-'+A, f'>{w}')
    else:
        if a+S<=w-2 and np.abs(x)>0.1:
            if c == 'ttl':
                return f"$ {S*'-'}{A}.{B[:w-S-1]} $"
            if c == 'csl':
                return format(f"{S*'-'}{A}.{B[:w-S-1-a]}", f'>{w}')
    u = len(str(e))
    if c == 'stm':
        z = 0
        q = len(str(m).strip('0').strip('.'))-2-S
        while z+1<=q and w>=S+1+(z+1)+1+len(str(e-(z+1))):
            z += 1
        return f"{m*10**z:1.0f}e{e-z}"
    if c == 'ttl':
        return fr"$ {format(m, f'1.{max(0, w-S-4-u)}f')} \times 10^{{ {e} }} $"
    if c == 'csl':
        return format(f"{format(m, f'1.{max(0, w-S-3-u)}f')}e{e}", f'>{w}')
    raise ValueError(f"unknown context: {c}")

@beartype
def unit(
    x: str,
    c: str = 'csl',
) -> str:
    """
    Return a notation of the unit x in context c.

    Input:
        x (str): unit
        c (str): context

    Output:
        s (str): notation of x
    """
    if c == 'ttl':
        return fr"$ \mathrm{{ {x} }} $"
    else:
        return x.replace("{", "").replace("}", "").replace("^", "")

@beartype
def quantity(
    v: Scalar,
    u: str,
    c: str = 'csl',
    w: int = 4,
) -> str:
    """
    Return the notation of value v and unit u in context c.

    Input:
        v (Scalar): value
        u (str): unit
        c (str): context
        w (int): width of the value

    Output:
        s (str): notation of u and v as a physical quantity
    """
    val = number(v, c, w)
    uni = unit(u, c)
    if c == 'ttl':
        val = val.replace(r"$", "").strip()
        uni = uni.replace(r"$", "").strip()
        return fr"$ {val} {uni} $"
    else:
        return val+uni

@beartype
def equality(
    a: str,
    b: str,
    c: str = 'csl',
) -> str:
    """
    Return the notation of a=b in context c.

    Input:
        a (str): left term of equality
        b (str): right term of equality
        c (str): context

    Output:
        s (str): notation of the equality
    """
    if c == 'ttl':
        a = a.replace(r"$", "").strip()
        b = b.replace(r"$", "").strip()
        return fr"$ {a} = {b} $"
    else:
        a = a.replace("\\", "")
        if c == 'csl':
            return a+"="+b
        elif c == 'stm':
            return a+b

@beartype
def parameters(
    r: dict,
    c: str = 'csl',
    s: bool = True,
) -> str:
    """
    Return the notation describing the model parameters set r.

    Input:
        r (dict): model parameters
        c (str): context
        s (bool): display the seed

    Output:
        n (str): model parameters description
    """
    # user-defined name for the set of parameters
    if 'name' in r:
        return r['name']
    n = ""
    # version
    if 'v' in r:
        n += "-"+r['v']
    # parameter list
    p = []
    if 'd' in r:  # density
        p.append(equality('d', quantity(r['d'], "nm^{-2}", c), c))
    if 's' in r: # subarea/cell side
        p.append(equality("s", number(r['s'], c, 4)+unit("nm", c), c))
    if 'f' in r: # filling
        p.append(equality("f", str(r['f']), c))
    if 't' in r: # wall thickness
        p.append(equality("t", number(r['t'], c, 3)+unit("nm", c), c))
    if 'l' in r: # dipole length
        p.append(equality("l", number(r['l'], c, 3)+unit("nm", c), c))
    # concatenate
    if len(p) > 0:
        if c == 'stm':
            n += "_"+"_".join(p)
        elif c == 'ttl':
            p = [q.replace("$", "").strip() for q in p]
            n += r" $ \left( "+r", \ ".join(p)+r" \right) $"
        elif c == 'csl':
            n += f" ({' '.join(p)})"
        else:
            raise ValueError(f"unknown context {c}")
    return n

@beartype
def fmt(
    f: str,
    v: dict,
    c: str = 'csl',
) -> str:
    """
    Return the elements of v selected and ordered by f.

    Input:
        f (str): format
        v (dict): dictionary containing the parameters
        c (str): context

    Output:
        s (str): elements of v selected and ordered by f
    """
    if c == 'stm':
        sep = '_'
    else:
        sep = " "
    return sep.join([v[n] for n in f if not v[n] is None])



#!/usr/bin/env python
# coding: utf-8

"""
Models for the random generation of dislocations.

The following fields can be used in the model parameters:
    v: variant (in RRDD or RCDD)
    d: density (in RDD, RRDD or RCDD)
    s: side of the subarea (in RRDD) or the cell (for RCDD)
    f: filling of the subareas (in RRDD or RCDD)
    t: wall thickness (in RCDD)
    l: dipole length (in RCDD)
    name: parameters set nickname (in RDD, RRDD or RCDD)
"""

import math
from . import *
from . import geometries

@beartype
def RDD(
    g: str,
    s: Scalar,
    a: Scalar,
    r: dict,
    G: np.random._generator.Generator,
) -> tuple:
    """
    Return the positions and Burgers vector generated with RDD model.

    "RDD" means: "Random Dislocation Distribution"

    The positions of the dislocations are uniformly distributed in the
    region of interest.

    The number of dislocations with a positive Burger vector is equal
    to the number of dislocations with a negative Burger vector.

    Input:
        g (str): geometry of the region of interest
        s (Scalar): size of the region of interest [nm]
        a (Scalar): area of region of interest [nm^2]
        r (dict): model parameters
        G (np.random._generator.Generator): random number generator

    Output:
        p (VectorList): dislocation positions [nm]
        b (ScalarList): dislocation Burgers vectors sense [1]

    The following parameters can be specified in r:
        'd' (Scalar): density of dislocations [nm^-2]

    Complexity:
        O( r['d'] * a )
    """
    nh = round(r['d']*a/2) # half of the total number of dislocations
    nt = 2*nh # total number of dislocations
    # dislocation Burgers vectors
    b = np.concatenate((np.ones(nh), -np.ones(nh)))
    # dislocation positions
    if g == 'circle':
        phi = 2*np.pi*G.random(nt)
        rad = s*np.sqrt(G.random(nt))
        p = np.stack((rad*np.cos(phi), rad*np.sin(phi)), axis=1)
    elif g == 'square':
        p = s*G.random([nt, 2])
    return p, b

@beartype
def ticks(
    g: str,
    l: Scalar,
    s: Scalar,
) -> ScalarList:
    """
    Standardize the construction of a grid of subareas or cells.

    Input:
        g (str): geometry
        l (Scalar): shape size [nm]
        s (Scalar): size of the step [nm]

    Output:
        t (ScalarList): ticks along an axis

    Input example:
        g = 'circle'
        l = 3
        s = 1

    Output example:
        t = np.array([-3, -2, -1, 0, 1, 2, 3])

    Complexity:
        O( l/s )
    """
    m = math.ceil(l/s)
    if np.abs(m*s-l)/l>0.0001:
        warnings.warn("the step is not a divisor of the length", Warning)
    if g == 'circle':
        return s*np.arange(-m, m+1) # [-2s, -s, 0, s, 2s] with m=2
    if g == 'square':
        return s*np.arange(m+1) # [0, s, 2s] with m=2

@beartype
def even_positions(
    t: ScalarList,
    f: int,
) -> ScalarListList:
    """
    Return the grid case lower left corner positions.

    Input:
        t (ScalarList): ticks along an axis
        f (int): number of points in each case

    Output:
        p (ScalarList): case lower left corner positions

    Input example:
        t = np.array([0, 1])
        f = 4

    Output example:
        p = np.array([
            [0, 0], [0, 0], [0, 0], [0, 0],
            [1, 0], [1, 0], [1, 0], [1, 0],
            [0, 1], [0, 1], [0, 1], [0, 1],
            [1, 1], [1, 1], [1, 1], [1, 1],
        ])

    """
    x = np.repeat(np.tile(t, len(t)), f)
    y = np.repeat(t, f*len(t))
    return np.stack((x,y), axis=1)

@beartype
def even_senses(
    t: ScalarList,
    f: int,
) -> ScalarList:
    """
    Return the evenly distributed Burgers senses for even_positions.

    Input:
        t (ScalarList): ticks along an axis
        f (int): number of points in each case

    Output:
        b (ScalarList): Burgers vector senses

    Input example:
        t = np.array([0, 1])
        f = 4

    Output example:
        b = [- - + + - - + + - - + + - - + +]
    """
    bp = np.ones(f//2, dtype=int)
    b = np.tile(np.concatenate((bp, -bp)), len(t)**2)
    return b

@beartype
def RRDD(
    g: str,
    s: Scalar,
    a: Scalar,
    r: dict,
    G: np.random._generator.Generator,
) -> tuple:
    """
    Return the positions and Burgers vector generated with RRDD model.

    "RRDD" means: "Restrictedly Random Dislocation Distribution"

    There is a fixed number of dislocations per subarea. These
    dislocations are uniformly distributed within the subareas.

    In the square geometry, the number of dislocations with a positive
    Burger vector is equal to the number of dislocations with a
    negative Burger vector. When the parameter 'v' is set to 'R', the
    Burgers vectors are randomly distributed. When it is set to 'E',
    the number of positive dislocations is equal to the number of
    negative dislocations in each subarea.

    Input:
        g (str): geometry of the region of interest
        s (Scalar): size of the region of interest [nm]
        a (Scalar): area of region of interest [nm^2]
        r (dict): model parameters
        G (np.random._generator.Generator): random number generator

    Output:
        p (VectorList): dislocation positions [nm]
        b (ScalarList): dislocation Burgers vectors sense [1]

    The following parameters can be specified in r:
        'v' (str): variant ('R' or 'E')
        's' (int): side of the subareas [nm]
        'f' (int): number of dislocations per subarea
        'd' (Scalar): density of dislocations [nm^-2]

    The parameter 'd' is considered only if the parameter 'f' is not
    specified.

    Complexity:
        O( r['f'] * s^2/r['s']^2 )
    """
    if 'f' in r:
        f = r['f'] # specified filling
    else:
        f = round(r['d']*r['s']**2) # deduced filling
    if f%2 != 0:
        raise ValueError("odd number of dislocations per subarea")
    t = ticks(g, s, r['s'])[:-1] # left or bottom location of subareas
    nt = f * len(t)**2 # total number of dislocations
    nh = nt//2 # half of the total number of dislocations
    # dislocation positions
    p = even_positions(t, f) + r['s']*G.random((nt, 2))
    # dislocation Burgers vector
    if r['v'] == 'R':
        b = G.permutation(np.concatenate((np.ones(nh), -np.ones(nh))))
    elif r['v'] == 'E':
        b = even_senses(t, f)
    # masking
    if g == 'circle':
        m = geometries.mask(g, s, p)
        p, b = p[m], b[m]
    return p, b

@beartype
def RCDD(
    g: str,
    s: Scalar,
    a: Scalar,
    r: dict,
    G: np.random._generator.Generator,
) -> tuple:
    """
    Return the positions and Burgers vector generated with RCDD model.

    "RCDD" means: "Random Cell Dislocation Distribution"

    There is a fixed number of dislocations per subarea. These
    dislocations are uniformly distributed within the cell walls.

    In the square geometry, the number of dislocations with a positive
    Burger vector is equal to the number of dislocations with a
    negative Burger vector. When the parameter 'v' is set to 'R', the
    Burgers vectors are randomly distributed. When it is set to 'E',
    the number of positive dislocations is equal to the number of
    negative dislocations in each cell. When it is set to 'D', the
    positive dislocations and negative dislocations are spaced by a
    distance defined with the parameter 'l'.

    Input:
        g (str): geometry of the region of interest
        s (Scalar): size of the region of interest [nm]
        a (Scalar): area of region of interest [nm^2]
        r (dict): model parameters
        G (np.random._generator.Generator): random number generator

    Output:
        p (VectorList): dislocation positions [nm]
        b (ScalarList): dislocation Burgers vectors sense [1]

    The following parameters can be specified in r:
        'v' (str): variant ('R', 'E' or 'D')
        's' (int): side of the subareas [nm]
        't' (int): thickness of the cell walls [nm]
        'l' (int): length of the dipoles [nm]
        'f' (int): number of dislocations per subarea
        'd' (Scalar): density of dislocations [nm^-2]

    Complexity:
        O( r['d'] * a )
    """
    if 'f' in r:
        f = r['f'] # specified filling
    else:
        f = round(r['d']*r['s']**2) # deduced filling
    f = 2*max(1, round(f/2))
    if r['t'] > r['s']/2:
        raise ValueError("inconsistent thickness and cell side")
    t = ticks(g, s, r['s'])[:-1] # left or bottom location of subareas
    nt = f * len(t)**2 # total number of dislocations
    nh = nt//2 # half of the total number of dislocations
    # dislocation positions
    if r['v'] == 'D':
        rp = G.random((nh, 2))
    else:
        rp = G.random((nt, 2))
    l1 = r['t']/2 # wall brick length 1
    l2 = r['s'] - l1 # wall brick length 2
    m = G.integers(4, size=len(rp))
    m2 = np.stack((m, m)).T
    rp = np.where(m2==0, np.array([l1,  0])+np.array([l2, l1])*rp, rp)
    rp = np.where(m2==1, np.array([ 0, l2])+np.array([l2, l1])*rp, rp)
    rp = np.where(m2==2, np.array([ 0,  0])+np.array([l1, l2])*rp, rp)
    rp = np.where(m2==3, np.array([l2, l1])+np.array([l1, l2])*rp, rp)
    if r['v'] == 'D':
        phi = 2*np.pi*G.random(nh)
        ux = r['l']/2 * np.cos(phi)
        uy = r['l']/2 * np.sin(phi)
        ux = np.stack((ux, -ux), axis=1).ravel()
        uy = np.stack((uy, -uy), axis=1).ravel()
        rp = np.repeat(rp, 2, axis=0) + np.stack((ux, uy), axis=1)
    p = even_positions(t, f) + rp
    # dislocation Burgers vector
    if r['v'] == 'R':
        b = G.permutation(np.concatenate((np.ones(nh), -np.ones(nh))))
    elif r['v'] == 'E':
        b = even_senses(t, f)
    elif r['v'] == 'D':
        b = np.tile((1, -1), nh)
    # masking
    if g == 'circle':
        m = geometries.mask(g, s, p)
        p, b = p[m], b[m]
    return p, b



#!/usr/bin/env python
# coding: utf-8

"""
Tools for representing distributions and samples of distributions.
"""

from . import *
from . import overlap
from . import notation
from . import boundaries
from . import geometries

class Distribution:
    """
    Represent a distribution of dislocations.

    When the geometry is of type 'circle' the characteristic size s is
    its radius. When it is of type 'square' s is its side.

    The dislocations can be of type 'screw' or type 'edge'. If the type
    is not specified, type 'screw' is chosen by default.

    For circle geometries, the parameter c can be set to 'ISD'. For
    square geometries, it can be set to 'PBC<r>' or 'GBB<r>' where <r>
    is the rank of replications of the region of interest beyond the
    boundaries. The dislocations outside the area of ​​interest are then
    added to the distribution.

    Attributes:
        g (str): geometry of the region of interest
        s (Scalar): size of the region of interest [nm]
        n (int): dimension of space in the region of interest
        v (Scalar): n-volume of the region of interest [nm^n]
        m (GenerationFunction): distribution model generation function
        r (dict): distribution model parameters
        t (str): dislocation type
        p (VectorList): dislocation positions [nm]
        b (ScalarList): dislocation Burgers vector senses [1]
        d (Scalar): dislocation density [nm^-n]
        i (Scalar): inter dislocation distance [nm]
        c (NoneType|str): boundary conditions
        S (NoneType|int): random seed
        G (np.random._generator.Generator): random number generator
    """

    @beartype
    def __init__(self,
        g: str,
        s: Scalar,
        m: GenerationFunction,
        r: dict,
        t: str = 'screw',
        c: Optional[str] = None,
        S: Optional[int] = None,
        G: Optional[np.random._generator.Generator] = None,
    ) -> None:
        """
        Initialize the distribution.

        Input:
            g (str): geometry of the region of interest
            s (Scalar): size of the region of interest [nm]
            m (GenerationFunction): distribution model generation function
            r (dict): distribution model parameters
            t (str): dislocation type
            c (NoneType|str): boundary conditions
            S (NoneType|int): random seed
            G (NoneType|np.random._generator.Generator): random gen.

        Complexity:
            O( complexity_of(m) )
        """
        # shape
        if s <= 0:
            raise ValueError(f"incorrect size: {s}")
        self.s = s # characteristic size of the region of interest [nm]
        self.g = g # geometry of the region of interest
        self.n, self.v = geometries.nvolume(g, s) # dimension and n-volume
        # model
        self.m = m # model generation function
        self.r = r.copy() # model parameters
        # dislocations
        self.t = t # dislocation type
        if not G is None:
            self.G = G
            self.S = None
        else:
            self.G = np.random.default_rng(S) # random generator
            self.S = S # random seed
        self.p, self.b = self.m(self.g, self.s, self.v, self.r, self.G)
        self.d = len(self)/self.v # density of dislocations [nm^-n]
        if self.d == 0:
            raise ValueError("the model gives a density equal to 0")
        self.i = 1/np.sqrt(self.d) # inter dislocation distance [nm]
        # boundary conditions
        if c:
            cp, cb = self.conditions(c)
            self.p = np.concatenate((self.p, cp))
            self.b = np.concatenate((self.b, cb))
        self.c = c # boundary conditions name

    @beartype
    def __repr__(self) -> str:
        """Return the eval-able representation of the distribution."""
        args = (
            repr(self.g), # geometry
            str(self.s), # size
            self.m.__qualname__, # model generation function
            repr(self.r), # model parameters
            repr(self.t), # dislocation type
            repr(self.c), # bondary conditions
            repr(self.S), # random seed
        )
        return f"{self.__class__.__name__}({', '.join(args)})"

    @beartype
    def __str__(self) -> str:
        """Return a regular description of the distribution."""
        return self.name()

    @beartype
    def __len__(self) -> int:
        """Return the number of dislocations in the distribution."""
        return len(self.b)

    @beartype
    def conditions(self,
        c: str,
    ) -> tuple:
        """
        Return the dislocations for boundary conditions c.

        Input:
            c (str): boundary conditions name

        Output:
            cp (VectorList): outer dislocation positions [nm]
            cb (ScalarList): outer dislocation Burgers vectors sense [1]
        """
        if hasattr(self, 'c') and self.c:
            raise ValueError("boundary conditions already applied")
        if c=='ISD' and self.g=='circle' and self.t=='screw':
            cp = boundaries.image_positions(self.s, self.p)
            cb = - self.b
        elif 'PBC' in c and self.g=='square':
            u = boundaries.replication_displacements(int(c[3:]), self.s)
            cp = np.concatenate([self.p + u[i] for i in range(len(u))])
            cb = np.tile(self.b, len(u))
        elif 'GBB' in c and self.g=='square':
            u = boundaries.replication_displacements(int(c[3:]), self.s)
            cp = np.empty((0,2), dtype=self.p.dtype)
            cb = np.array([], dtype=self.b.dtype)
            for i in range(len(u)):
                p, b = self.m(self.g, self.s, self.v, self.r, self.G)
                cp = np.concatenate((cp, p + u[i]))
                cb = np.concatenate((cb, b))
        else:
            raise Exception(f"invalid boundary conditions: {c}")
        return cp, cb

    @beartype
    def name(self,
        f: str = 'dgsmtcS',
        c: str = 'csl',
    ) -> str:
        """
        Return a string describing the distribution in context c.

        Input:
            f (str): format
            c (str): context

        Output:
            csl|stm|ttl (str): string describing the distribution

        The format is defined by an arrangement of the characters:
            d: dislocation density [m^-2]
            g: geometry of the region of interest
            s: size of the region of interest [nm]
            m: model and its parameters
            t: dislocation type
            c: boundary conditions
            n: number of dislocations
            S: random seed
        """
        rho = notation.quantity(self.d*1e9**self.n, f"m^{{-{self.n}}}", c)
        v = {
            'd': notation.equality(r"\rho", rho, c),
            'g': self.g,
            's': notation.quantity(self.s, r"nm", c, w=4),
            'm': self.m.__name__+notation.parameters(self.r, c),
            't': self.t,
            'c': self.c,
            'n': str(int(len(self))),
            'S': notation.equality(r"S", str(self.S), c)
                if not self.S is None else None,
        }
        return notation.fmt(f, v, c)

    @beartype
    def w(self,
        a: Vector,
        r: ScalarList,
        r2: ScalarList,
    ) -> ScalarList:
        """
        Return the edge correction coefficients of the shape.

        Input:
            a (Vector): position of the neighborhoods center [nm]
            r (ScalarList): radius of the neighborhoods [nm]
            r2 (ScalarList): squared radius of neighborhoods [nm^2]

        Output:
            w (ScalarList): weighting coefficients for each radius

        Input example:
            a = np.array([x, y])
            r = np.array([r_0, r_1, r_2, ...])
            r2 = np.array([r_0^2, r_1^2, r_2^2, ...])

        Output example:
            w = np.array([w_r_0, w_r_1, w_r_2, ...])

        Complexity:
            O( len(r) )
        """
        # vo: n-volumes of the overlappings (for each value of r)
        # vv: n-volumes of the vicinities of a (for each value of r)
        if self.g == 'circle':
            d2 = np.sum(np.square(a)) # squared distance to the origin of a
            d = np.sqrt(d2) # distance to the origin of a
            vo = overlap.circle_circle(r, self.s, d, r2, self.s**2, d2)
            vv = np.pi*r2
        elif self.g == 'square':
            vo = overlap.circle_square(a[0], a[1], r, r2, self.s)
            vv = np.pi*r2
        w = np.divide(vo, vv, np.ones(vo.size), where=r2>0) # ratio
        return w

class Sample:
    """
    Represent a sample of distributions.

    Most of the attributes have a common meaning with those of class
    Distribution. The density and the inter dislocation distance are
    averaged over all the generated distributions.

    Attributes:
        l (tuple): generated distributions
        g (str): geometry of the region of interest
        s (Scalar): size of the region of interest [nm]
        n (int): dimension of space in the region of interest
        v (Scalar): n-volume of the region of interest [nm^n]
        m (GenerationFunction): distribution model generation function
        r (dict): distribution model parameters
        t (str): dislocation type
        d (Scalar): averaged dislocation density [nm^-n]
        i (Scalar): averaged inter dislocation distance [nm]
        c (NoneType|str): boundary conditions
        S (NoneType|int): random seed
        G (np.random._generator.Generator): random generator
    """

    @beartype
    def __init__(self,
        n: int,
        g: str,
        s: Scalar,
        m: GenerationFunction,
        r: dict,
        t: str = 'screw',
        c: Optional[str] = None,
        S: Optional[int] = None,
    ) -> None:
        """
        Initialize the sample of distributions.

        Input:
            n (int): number of distributions to generate
            g (str): geometry of the region of interest
            s (Scalar): size of the region of interest [nm]
            m (GenerationFunction): model generation function
            r (dict): model parameters
            t (str): dislocation type
            c (NoneType|str): boundary conditions
            S (NoneType|str): random seed

        Complexity:
            O( c * complexity_of(Distribution) )
        """

        self.G = np.random.default_rng(S) # random generator
        self.S = S # random seed
        args = (g, s, m, r, t, c, None, self.G) # distribution arguments
        if n <= 0:
            raise ValueError(f"incorrect number of distribution: {n}")
        self.l = tuple([Distribution(*args) for i in range(n)])
        self.g = g # geometry of the region of interest
        self.s = s # size of the region of interest [nm]
        self.n = self[0].n # dimension of space
        self.v = self[0].v # n-volume of the region of interest [nm^n]
        self.m = m # model function
        self.r = r # model parameters
        self.t = t # dislocation type
        self.d = self.average(lambda d : d.d) # averaged dislocation density
        self.i = self.average(lambda d : d.i) # averaged inter disl. dist.
        self.c = c # boundary conditions

    @beartype
    def __repr__(self) -> str:
        """Return the eval-able representation of the sample."""
        args = (
            str(len(self)), # number of distributions
            repr(self.g), # geometry
            str(self.s), # size
            self.m.__qualname__, # model function
            repr(self.r), # model parameters
            repr(self.t), # dislocation type
            repr(self.c), # boundary conditions
            repr(self.S), # random seed
        )
        return f"{self.__class__.__name__}({', '.join(args)})"

    @beartype
    def __str__(self) -> str:
        """Return a regular description of the sample."""
        return self.name()

    @beartype
    def __getitem__(self, k: int) -> Distribution:
        """Return the k-th distribution stored in the sample."""
        return self.l[k]

    @beartype
    def __len__(self) -> int:
        """Return the number of distributions stored in the sample."""
        return len(self.l)

    @beartype
    def name(self,
        f: str = 'ndgsmtcS',
        c: str = 'csl',
    ) -> str:
        """
        Return a string describing the sample in context c.

        Input:
            f (str): format
            c (str): context

        Output:
            csl|stm|ttl (str): string describing the sample

        The format is defined by an arrangement of the characters:
            d: dislocation density [m^-2]
            g: geometry of the region of interest
            s: size of the region of interest [nm]
            m: model and its parameters
            t: dislocation type
            c: boundary conditions
            n: number of distributions generated
            S: random seed
        """
        rho = notation.quantity(self.d*1e9**self.n, f"m^{{-{self.n}}}", c)
        v = {
            'd': notation.equality(r"\rho", rho, c),
            'g': self.g,
            's': notation.quantity(self.s, r"nm", c, w=4),
            'm': self.m.__name__+notation.parameters(self.r, c),
            't': self.t,
            'c': self.c,
            'n': str(int(len(self))),
            'S': notation.equality(r"S", str(self.S), c)
                if not self.S is None else None,
        }
        return notation.fmt(f, v, c)

    @beartype
    def average(self,
        f: Callable[[Distribution, Any], AnalysisOutput],
        *args,
    ) -> AnalysisOutput:
        """
        Average the function f over the set of generated distributions.

        The first parameter of f must be the distribution to analyze.
        The function f must return a summable object or a tuple of
        summable objects. In the case of tuple, the items are averaged
        one by one according to their position in the tuple.

        Input:
            f (Callable): function applied to a distribution
            *args: additional arguments to pass to function f

        Output:
            r (AnalysisOutput): result of f averaged over the distrib.

        Complexity:
            O( len(self) * complexity_of(f) )
        """
        r = f(self[0], *args) # result of f on the first distribution
        if isinstance(r, tuple):
            r = list(r)
            n = len(r)
            for i in range(1, len(self)):
                ri = f(self[i], *args) # result of f on the ith distribution
                for j in range(n):
                    r[j] += ri[j]
            for j in range(n):
                r[j] = r[j]/len(self)
            return tuple(r)
        else:
            for i in range(1, len(self)):
                r += f(self[i], *args)
            return r/len(self)



#!/usr/bin/env python
# coding: utf-8

"""
Standardized files for input to an X-ray diffraction simulation program.
"""

import os
from . import *
from . import sets

@beartype
def contrast_factor(
    t: str,
    g: Vector,
    l: Vector,
    b: Vector,
    nu: Scalar,
) -> Scalar:
    """
    Return the contrast factor in elastically isotropic crystals.

    Input:
        t (str): dislocation type
        g (Vector): diffraction vector direction (hkl)
        l (Vector): dislocation line vector direction [uvw]
        b (Vector): Burgers vector direction [uvw]
        nu (Scalar): Poisson's number [1]

    Output:
        C (Scalar): dislocation contrast factor [1]

    Input example:
        t = 'screw'
        g = np.array([2, 0, 0])
        l = np.array([1, 1, 0])
        b = np.array([1, 1, 0])
        nu = 0.345

    Output example:
        C = 0.25
    """
    ng = np.linalg.norm(g) # diffraction vector norm to normalize
    nl = np.linalg.norm(l) # dislocation line vector norm to normalize
    psi = np.arccos(np.dot(g, l)/(ng*nl)) # angle between g and l
    if t == 'screw':
        C = np.sin(psi)**2 * np.cos(psi)**2
    elif t == 'edge':
        pg = g - (np.dot(g, l)/nl**2)*l # projection of g
        pb = b - (np.dot(b, l)/nl**2)*l # projection of b
        npg = np.linalg.norm(pg) # norm of the projection of g
        npb = np.linalg.norm(pb) # norm of the projection of b
        gamma = np.arccos(np.dot(pg, pb)/(npg*npb)) # angle betw. projections
        C = (np.sin(psi)**4
            / (8*(1-nu)**2)
            * (1-4*nu+8*nu**2+4*(1-2*nu)*np.cos(gamma)**2))
    return C

dft_l = {'edge': np.array([ 1, -1, -2]), 'screw': np.array([ 1,  1,  0])}
dft_L = {'edge': np.array([ 1,  1,  0]), 'screw': np.array([-1,  1,  0])}

@beartype
def export_distribution(
    d: sets.Distribution,
    i: Scalar,
    **kwargs,
) -> None:
    """
    Export the dislocations of d to a standardized input data file.

    The inter-dislocation distance must be specified because it may be
    different from d.i when averaged over several distributions.

    Input:
        d (Distribution): distribution to be exported
        i (Scalar): inter dislocation distance [nm]
      **g (Vector): diffraction vector direction (hkl) (default: 2,0,0)
      **b (Vector): Burgers vector direction [uvw] (default: 2,0,0)
      **l (Vector): line vector direction [uvw] (default: see d. type)
      **L (Vector): direction along Lx [uvw] (defaut: see d. type)
      **a (Scalar): cell side [nm] (default: 0.40494)
      **a3 (Scalar): step size along Lx [nm] (default: depends on i)
      **nu (Scalar): Poisson's number [1] (default: 0.345)
      **pbc (int): replications rank (default: 0)
      **expdir (str): export directory (default: '')
      **expfmt (str): export format (default: 'dat')
      **expstm (str): export stem (default: d.name())

    Complexity:
        O( len(d) )
    """
    # optional parameters
    pbc = getkwa('pbc', kwargs, int, 0)
    if pbc > 0:
        stmpbc = '_PBC'+str(pbc)
    else:
        stmpbc = ''
    g = getkwa('g', kwargs, Vector, np.array([2, 0, 0]))
    b = getkwa('b', kwargs, Vector, np.array([1, 1, 0]))
    l = getkwa('l', kwargs, Vector, dft_l[d.t])
    L = getkwa('L', kwargs, Vector, dft_L[d.t])
    a = getkwa('a', kwargs, Scalar, 0.40494)
    a3 = getkwa('a3', kwargs, Scalar, max(2, i/12))
    nu = getkwa('nu', kwargs, Scalar, 0.345)
    expdir = getkwa('expdir', kwargs, str, '')
    expfmt = getkwa('expfmt', kwargs, str, 'dat')
    expstm = getkwa('expstm', kwargs, str, d.name(c='stm')+stmpbc)
    endkwa(kwargs)
    # parameters
    if d.t=='screw' and np.linalg.norm(np.cross(l, b))!=0:
        raise ValueError("screw type but l and b not parallel")
    elif d.t=='edge' and np.dot(l, b)!=0:
        raise ValueError("edge type but l and b not perpendicular")
    C = contrast_factor(d.t, g, l, b, nu)
    if d.g == 'circle':
        str_g = "Cylinder radius"
    if d.g == 'square':
        if pbc > 0:
            str_g = f"Square_{pbc} side"
        else:
            str_g = "Square side"
    # write
    indices = lambda v: " ".join([format(c, '2.0f') for c in v])
    with open(os.path.join(expdir, expstm+"."+expfmt), "w") as f:
        h = (f"# please keep the structure of this file unchanged\n"
             f"{indices(l)} # z: direction of 'l' (line vector) [uvw]\n"
             f"{indices(L)} # x: direction of 'L' (Fourier variable) [uvw]\n"
             f"{indices(b)} # b: Burgers vector direction [uvw]\n"
             f"{indices(g)} # g: diffraction vector direction (hkl)\n"
             f"{C:8.6f} # C: contrast coefficient [1]\n"
             f"{a:8.6f} # a: cell parameter [nm]\n"
             f"{d.s:8.0f} # s: {str_g} [nm]\n"
             f"{a3:8.1f} # a3: step size of 'L' along x [nm]\n"
             f"{nu:8.3f} # nu: Poisson's number [1]\n"
             f"{len(d):8.0f} # number of dislocations\n"
             f"# Burgers vector and dislocation (x,y) coordinates\n")
        f.write(h)
        fmt = "%2.0f %22.15E %22.15E"
        np.savetxt(f, np.stack((d.b, d.p[:,0], d.p[:,1])).T, fmt=fmt)

@beartype
def export_sample(
    s: sets.Sample,
    **kwargs,
) -> None:
    """
    Export a standardized input data file for each distribution in s.

    Input:
        s (Sample): sample of distributions to be exported
        <see export_distribution function for keyword arguments>

    Complexity:
        O( len(s) * complexity_of(export_distribution) )
    """
    # optional parameters
    pbc = getkwa('pbc', kwargs, int, 0)
    if pbc > 0:
        stmpbc = '_PBC'+str(pbc)
    else:
        stmpbc = ''
    expdir = getkwa('expdir', kwargs, str, '')
    expstm = getkwa('expstm', kwargs, str, s.name(c='stm')+stmpbc)
    # export
    stmdir = os.path.join(expdir, expstm) # folder where to export the files
    w = len(str(len(s))) # number of characters in file names
    if os.path.exists(stmdir): # if the sample already exists
        for f in os.listdir(stmdir): # browse the distributions
            os.remove(os.path.join(stmdir, f)) # delete the existing
    else: # if the sample does not exists
        os.mkdir(stmdir) # create the folder
    for i in range(len(s)): # export the distribution files
        export_distribution(
            s[i], # ith distribution
            s.i, # average interdislocation distance
            expdir=stmdir, # export directory
            expstm=str(i+1).zfill(w), # export stem
            pbc=pbc, # periodic boundary conditions
            **kwargs,
        )

@beartype
def export(
    o: Union[sets.Distribution, sets.Sample],
    **kwargs,
) -> None:
    """
    Versatile function for exporting a standardized data file.

    Input:
        o (Distribution|Sample): distribution or sample to export
        <see export_distribution function for keyword arguments>

    Complexity:
        O( complexity_of(export_distribution) ) if o is a distribution
        O( complexity_of(export_sample) ) if o is a sample
    """
    if isinstance(o, sets.Distribution):
        export_distribution(o, o.i,**kwargs)
    else:
        export_sample(o, **kwargs)
